#ifndef CONTEXT
#error "CONTEXT undefined"
#endif
#ifndef PARSE
#error "PARSE undefined"
#endif
#ifndef HEADER
#error "HEADER undefined"
#endif
#ifndef CLAUSE
#error "CLAUSE undefined"
#endif
#ifndef BINARY
#define "BINARY undefined"
#endif

static void PARSE () {
  int prev = 0, deleted = 0, header = 0, expected_maxidx = INT_MAX;
  int parsed_clauses = 0, expected_clauses = 0, header_warning = 0;
  const int expect_header = CONTEXT.type;
  unsigned long saved_charno;
#if MAXVERBOSITY > 0
  Timer timer;
  start_timer (1, &timer);
#endif
  current_context = &CONTEXT;
  open_input_file_for_reading (CONTEXT.read_from_stdin ? 0 : CONTEXT.name);
  msg1 ("");
  msg1 ("reading %s from '%s'", context_type (&CONTEXT), input_file_name);
  lineno = charno = end_of_line = readchars = 0;
  if (BINARY) goto DONE;
  for (;;) {
    int ch = read_next_char ();
    if (is_white_space (ch)) continue;
    start_of_token = charno;
    if (ch == EOF) {
      if (deleted) perr ("'d' before end-of-file without literals");
      if (prev) perr ("terminating zero missing at end-of-file");
      break;
    }
    if (ch == 'c') {
      while ((ch = read_next_char ()) != '\n')
	if (ch == EOF) perr ("unexpected end-of-file in comment");
    } else if (!expect_header && ch == 's') {
      if (charno > 1) perr ("'s' not at the beginning of a line");
      while ((ch = read_next_char ()) != '\n')
	if (ch == EOF) perr ("unexpected end-of-file in solution line");
      if (!EMPTY (line))
	msg1 ("found '%s' in '%s'", line.start, input_file_name);
      if (strcmp (line.start, "s SATISFIABLE") &&
          strcmp (line.start, "s UNSATISFIABLE") &&
          strcmp (line.start, "s UNKNOWN")) {
	assert (charno > 1), charno--;
	pwrn ("invalid 's' line");
      }
    } else if (!expect_header && ch == 'v') {
      // TODO partially parse 'v' lines ...
      if (charno > 1) perr ("'v' not at the beginning of a line");
      while ((ch = read_next_char ()) != '\n')
	if (ch == EOF) perr ("unexpected end-of-file in witness line");
    } else if (ch == 'p') {
      saved_charno = charno;
      assert (charno > 0);
      if (charno > 1) perr ("'p' not at the beginning of a line");
      if (header) perr ("unexpected second 'p'");
      if ((ch = read_next_char ()) != ' ')
	perr ("expected space after 'p'%s", but_got (ch));
      if ((ch = read_next_char ()) != 'c')
	perr ("expected 'c' after 'p '%s", but_got (ch));
      if ((ch = read_next_char ()) != 'n')
	perr ("expected 'n' after 'p c'%s", but_got (ch));
      if ((ch = read_next_char ()) != 'f')
	perr ("expected 'f' after 'p cn'%s", but_got (ch));
      if ((ch = read_next_char ()) != ' ')
	perr ("expected space after 'p cnf'%s", but_got (ch));
      start_of_token = charno + 1;
      expected_maxidx = parse_non_negative_int ("maximum variable index");
      ch = read_next_char ();
      start_of_token = saved_charno;
      if (ch != ' ')
	perr ("expected space after 'p cnf %d'%s",
	  expected_maxidx, but_got (ch));
      start_of_token = charno + 1;
      expected_clauses = parse_non_negative_int ("number of clauses");
      start_of_token = saved_charno;
      while ((ch = read_next_char (ch)) != '\n')
	if (!is_white_space (ch))
	  perr (
	    "expected white-space or new line after header 'p cnf %d %d'%s",
	    expected_maxidx, expected_clauses, but_got (ch));
      HEADER (expected_maxidx, expected_clauses);
      header = 1;
    } else if (ch == 'd') {
      if (deleted) perr ("'d' follows 'd' without literal");
      if (charno > 1) perr ("'d' not at the beginning of a line");
      if (prev) perr ("terminating zero missing before 'd'");
      position_at_start_of_clause = lineno;
      ch = read_next_char ();
      if (!is_white_space (ch))
	perr ("expected white-space after 'd'%s", but_got (ch));
      deleted = 1;
    } else {
      int sign = 1, idx, lit;
      if (!deleted) position_at_start_of_clause = lineno;
      if (ch == '-') {
	ch = read_next_char ();
	if (!isdigit (ch))
	  perr ("expected digit after '-'%s", but_got (ch));
	if (ch == '0')
	  perr ("expected non-zero digit after '-'%s", but_got (ch));
	sign = -1;
      } else if (!isdigit (ch)) {
	if (deleted)
	  perr ("expected '-' or digit after 'd'%s", but_got (ch));
	else if (header)
	  perr ("expected 'c', 'd', '-' or digit%s", but_got (ch));
	else
	  perr ("expected 'p', 'c', 'd', '-' or digit%s", but_got (ch));
      }
      save_char (ch);
      idx = parse_non_negative_int ("variable index");
      if (!is_white_space (ch = read_next_char ())) {
	if (idx) {
	  assert (!(sign == -1 && idx == INT_MIN));
	  perr ("expected white-space after literal '%d'%s",
	    sign*idx, but_got (ch));
	} else perr ("expected white-space after '0'%s", but_got (ch));
      }
      save_char (ch);
      if (expect_header && !header && !header_warning)
	pwrn ("header 'p cnf <maxvar> <clauses>' expected"),
	header_warning = 1;
      if (header && idx > expected_maxidx)
	perr ("variable index %d exceeds maximum %d", idx, expected_maxidx);
      if (header && parsed_clauses == expected_clauses)
	perr ("too many clauses (expected %d)", expected_clauses);
      lit = sign * idx;
      if (lit) {
	if (COUNT (clause) == (long) UINT_MAX)
	  perr ("clause with more than UINT_MAX = %u literals", UINT_MAX);
	PUSH (clause, lit);
      } else {
	CLAUSE (deleted);
	CLEAR (clause);
	deleted = 0;
	parsed_clauses++;
      }
      prev = lit;
    }
  }
DONE:
  close_input_file_if_necessary ();
#if MAXVERBOSITY > 0
  stop_timer (&timer,
     "parsed %lu clauses from '%s' in",
     parsed_clauses, input_file_name);
#endif
  if (header && parsed_clauses < expected_clauses ) {
    int missing = expected_clauses - parsed_clauses;
    if (missing == 1) perr ("one clause missing");
    else perr ("%d clauses missing", missing);
  } else if (!header && expect_header && !header_warning) 
    pwrn ("header 'p cnf <maxvar> <clauses>' missing");
}

#undef CONTEXT
#undef PARSE
#undef CLAUSE
#undef HEADER
#undef BINARY
